#include <vector>
#include <ProbabilisticModel.h>

class Run
{

public:
	
	void record_errors (double);
	void record_previous_errors ();
	bool export_run_errors(std::ofstream &) const;
	
	friend class Experiment;
		
private:
		
	std::vector < double > errors;
};

// The following will contain data necessary to do batch jobs

class Experiment
{
	
public:
		
	// Constructor
		
	Experiment ();
	
	/*
	void run(const std::string, const std::string);
	
	void export_errors(const std::string output_file_name);
	
	void initialize();
	 */
	
	// Normal methods
		
	void get_inference();
	
	void record_errors();
	void callback_monitoring (const double);
	void set_running_time (const unsigned int);
	void set_burnoff (const unsigned int);
	void set_running_particles (const unsigned int);
	void set_ground_truth_method (const std::string);
	void set_timer_method (const bool);
	
	// Import/Export methods
	
	bool set_up_model_from_file(const std::string &);
	bool export_results_to_file(const std::string &);
	
	bool serialize_errors(std::ofstream &);
	
	bool unserialize_method(std::ifstream &);
	
	enum 
	{ 
		exact_belief_propagation,
		loopy_belief_propagation,
		simple_gibbs_sampling,
		tree_gibbs_sampling,
		tree_partition
	};
		
private:
		
	ProbabilisticModel * model;
	
	//bool monitoring; // monitoring is now a global boolean
	
	unsigned int inference_method;
	unsigned int steps;
	unsigned int burnoff;
	bool using_rao_blackwell;
	bool using_timer;
	double sampling_interval;
	double next_sample_time;	

	std::vector <Run> runs;
	Run * current_run;
	unsigned int number_runs;
	
	bool only_partition;
	bool strict_error_timing;

	/*
	bool np_tree_gibbs;
	bool metropolis_bool;
	bool ground_metropolis;
	bool ground_tree;
	bool ground_exact;
	 */
		
};
