// This file will be used to implement the tree Partition

#include <iostream>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/graph_traits.hpp>
#include <boost/graph/subgraph.hpp>
#include <boost/graph/copy.hpp>
#include <queue>

#include <GraphTreePartition.h>
#include <Potential.h>
#include <RandomVariable.h>
#include <Node.h>
#include <Generic.h>
#include <OutputGraph.h>

using namespace std;
using namespace boost;

// Explicit template instantiations

typedef property< edge_index_t, unsigned int, DiscretePotential *> DiscretePairwiseEdgeProperty;
typedef adjacency_list<vecS, vecS, undirectedS, MessageNode <DiscreteRandomVariable, boost::adjacency_list_traits<vecS, vecS, undirectedS>::vertex_descriptor> *, DiscretePairwiseEdgeProperty > DiscretePairwiseGraph;
typedef subgraph <DiscretePairwiseGraph> SubDiscretePairwiseGraph;

typedef property< edge_index_t, unsigned int> EdgeProperty;
typedef adjacency_list<vecS, vecS, undirectedS, MessageNode <DiscreteRandomVariable, boost::adjacency_list_traits<vecS, vecS, undirectedS>::vertex_descriptor> *, EdgeProperty > DiscreteFactorGraph;
typedef subgraph <DiscreteFactorGraph> SubDiscreteFactorGraph;

typedef property< edge_index_t, unsigned int, ContinuousPotential *> NPEdgeProperty;
typedef adjacency_list<vecS, vecS, undirectedS, GraphicalModelNode *, NPEdgeProperty > NPGraph;
typedef subgraph <NPGraph> SubNPGraph;

// End of the typedefs

template void partition_graph_into_trees< SubDiscretePairwiseGraph > ( SubDiscretePairwiseGraph &);
template void partition_graph_into_trees< SubDiscreteFactorGraph > (SubDiscreteFactorGraph &);

template void partition_mrf_graph_into_chains < subgraph <NPGraph> > (subgraph <NPGraph> &);
template void partition_mrf_graph_into_chains < subgraph <DiscretePairwiseGraph> > (subgraph <DiscretePairwiseGraph> &);

template void partition_simple_chain < subgraph <NPGraph> > (subgraph <NPGraph> &);
template void partition_simple_chain < SubDiscretePairwiseGraph > (SubDiscretePairwiseGraph &);

// End of Explicit template instantiations

//********** Implementation of GraphPartitionQueueOrder ****************//

template <class Graph, class VertexDescriptor>
GraphPartitionQueueOrder<Graph, VertexDescriptor>::GraphPartitionQueueOrder(const Graph & a) : g(a)
{
}

// This is the comparison function. If it returns "true" it means t_1 will be LATER in the queue, ie t_1 > t_2
// If we want it to.

template <class Graph, class VertexDescriptor>
bool GraphPartitionQueueOrder<Graph, VertexDescriptor>::operator()(myTuple t_1, myTuple t_2) const
{
	// 1: choices
	// 2: overall degree
	// 3: step
	// 4: random
	
	if ( tuples::get<1>(t_1)  == tuples::get<1>(t_2) ) // If the number of choices is the same
	{
		
		if ( tuples::get<2>(t_1) == tuples::get<2>(t_2) ) // If the overall degree is the same	
		{
			
			if ( tuples::get<3>(t_1) == tuples::get<3>(t_2) ) // If the steps are the same
			{
				
				//return ( tuples::get<4>(t_1) < tuples::get<4>(t_2) ); // quantity representing the desirability of going into that vertex
				
				//at random
				
				return (uniform_distribution_0_1() < 0.5);
			}
			
			else // we decide by the steps
			{
				return ( tuples::get<3>(t_1) < tuples::get<3>(t_2) ); 
			}
			
		}
		
		else // we decide by the lowest overall degree
		{
			return ( tuples::get<2>(t_1) > tuples::get<2>(t_2) );
		}
		
	}
	
	else // we decide by the number of choices
	{
		return ( tuples::get<1>(t_1)  > tuples::get<1>(t_2) ); 
	}
}

/*
 template <class Graph, class VertexDescriptor>
 bool GraphPartitionQueueOrder<Graph, VertexDescriptor>::operator()(myTuple t_1, myTuple t_2) const
 {
	 // Totally random
	 
	 return (uniform_distribution_0_1() < 0.5);
 }
 */

//********** End of Implementation of GraphPartitionQueueOrder ****************//

template <class Graph, class ColorMap>
unsigned int compute_possible_choices(const Graph & g, ColorMap & color, const typename graph_traits<Graph>::vertex_descriptor v)
{
	unsigned int a(0);
	
	typename graph_traits<Graph>::adjacency_iterator u, u_end;
	typedef typename property_traits<ColorMap>::value_type ColorValue;
	typedef color_traits<ColorValue> Color;
	
	for (tie (u, u_end) = adjacent_vertices(v, g); u != u_end; ++u)
	{
		if (get(color, *u) == Color::white())
		{
			++a;
		}
	}
	
	return a;
}

template <class Graph, class ColorMap>
bool prune_out_degree_0 (Graph & g, ColorMap color)
{
	typedef typename property_traits<ColorMap>::value_type ColorValue;
	typedef color_traits<ColorValue> Color;
	typename graph_traits <Graph>::vertex_iterator u, u_end;
	
	for (tie(u, u_end) = vertices(g); u != u_end; ++u)
	{
		if ( out_degree(*u, g) == 0)
		{
			put (color, *u, Color::red() );
			return true;
		}
	}
	
	return false;
}


template <class Graph, class ColorMap, class ParentMap>
void prune_out_degree_1 (Graph & g, ColorMap color, ParentMap parent)
{
	bool prune_again = true;
	
	typedef typename property_traits<ColorMap>::value_type ColorValue;
	typedef color_traits<ColorValue> Color;
	typename graph_traits <Graph>::vertex_iterator u, u_end,v, v_end;
	typename graph_traits <Graph>::vertex_descriptor w;
	
	while (prune_again)
	{
		prune_again = false;
		
		for (tie(u, u_end) = vertices(g); u != u_end; ++u)
		{
			
			if ( out_degree(*u, g) == 1)
			{	
				w = *(adjacent_vertices(*u,g).first);
				clear_vertex(*u,g);
				put(color, *u, Color::green());
				prune_again = true;
				
				// Need to record parent, that should do it
				
				for (tie(v, v_end) = vertices(g); v != v_end; ++v)
				{
					if (get(parent,*v) == *u)
					{
						put (parent, *v, w);
					}
				}
			}
		}
	}
}

template <class Graph, class ColorMap, class ParentMap>
void prune_out_degree_2 (Graph & g, ColorMap color, ParentMap parent)
{	
	typedef typename property_traits<ColorMap>::value_type ColorValue;
	typedef color_traits<ColorValue> Color;
	typedef typename graph_traits <Graph>::vertex_descriptor VertexDescriptor;
	
	typename graph_traits <Graph>::vertex_iterator u, u_end, x, x_end;
	typename graph_traits <Graph>::adjacency_iterator v, v_end, w;
	
	for (tie(u, u_end) = vertices(g); u != u_end; ++u)
	{
		
		if ( out_degree(*u, g) == 2 ) // we also have to check if there isn't an edge connecting the two neighbours
		{
			tie (v,v_end) = adjacent_vertices (*u,g);
			w = v;
			++w;
			
			VertexDescriptor a = *v;
			VertexDescriptor b = *w;
			
			if (edge(*v, *w, g).second) // this means there is an edge between v and w
			{
				continue;
			}
			
			//cout << "pruned " << *u+1 << endl;
			
			remove_edge(*u, a, g);
			
			remove_edge(*u, b, g);
			
			add_edge(a,b,g);
			
			
			put(color, *u, Color::green());
			
			
			// Need to record parent
			
			for (tie(x, x_end) = vertices(g); x != x_end; ++x)
			{
				if (get(parent,*x) == *u)
				{
					put (parent, *x, *w);
				}
			}
			
		}
		
	}
}



template <class Graph, class ColorMap>
bool check_black_neighbor(Graph & g, ColorMap & color, const typename graph_traits<Graph>::vertex_descriptor v)
{
	typename graph_traits<Graph>::adjacency_iterator u, u_end;
	typedef typename property_traits<ColorMap>::value_type ColorValue;
	typedef color_traits<ColorValue> Color;
	
	for (tie (u, u_end) = adjacent_vertices(v, g); u != u_end; ++u)
	{
		if (get(color, *u) == Color::black())
		{
			return true;
		}
	}
	
	return false;
}

template <class Graph, class ColorMap>
bool check_gray_neighbor(Graph & g, ColorMap & color, const typename graph_traits<Graph>::vertex_descriptor v)
{
	typename graph_traits<Graph>::adjacency_iterator u, u_end;
	typedef typename property_traits<ColorMap>::value_type ColorValue;
	typedef color_traits<ColorValue> Color;
	
	for (tie (u, u_end) = adjacent_vertices(v, g); u != u_end; ++u)
	{
		if (get(color, *u) == Color::gray())
		{
			return true;
		}
	}
	
	return false;
}

template <class Graph, class ColorMap>
void transmit_escape(typename graph_traits<Graph>::vertex_descriptor v, ColorMap & color, vector <unsigned int> & free_edges, vector<list <typename graph_traits<Graph>::vertex_descriptor> > & available_neighbors, Graph & g)
{
	//typename graph_traits<Graph>::adjacency_iterator u, u_end;
	
	typedef typename property_traits<ColorMap>::value_type ColorValue;
	typedef color_traits<ColorValue> Color;
	
	typename graph_traits<Graph>::vertex_descriptor u = available_neighbors[v].front();
	
	//cout << "We decrease vertex " << g[u]->get_random_variable()->get_index() << " which now has " <<  free_edges [u] -1 << endl;
	
	free_edges [u] = free_edges [u] -1;
	available_neighbors[u].remove(v);
	
	if (free_edges [u] == 1)
	{
		transmit_escape(u, color, free_edges, available_neighbors, g);
	}
	
	
	/*
	 for (tie (u, u_end) = adjacent_vertices(v, g); u != u_end; ++u)
	 {
		 // We are not interested in propagating if neighbors are red or black.
		 
		 if ( (get(color, *u) != Color::red()) && (get(color, *u) != Color::black() ))
		 {
			 cout << "We decrease vertex " << *u+1 << " which now has " <<  free_edges [*u] -1 << endl;
			 
			 free_edges [*u] = free_edges [*u] -1;
			 
			 if (free_edges [*u] == 1)
			 {
				 transmit_escape(*u, color, free_edges, g);
			 }
		 }
	 }
	 */
}

//template <class Graph, class ColorMap, class ParentMap>



typedef iterator_property_map < vector < default_color_type >::iterator, property_map < SubDiscretePairwiseGraph, vertex_index_t>::type > DiscretePairwiseGraphColorMap;
typedef iterator_property_map < vector < graph_traits <SubDiscreteFactorGraph>::vertex_descriptor >::iterator, property_map < SubDiscretePairwiseGraph, vertex_index_t>::type > DiscretePairwiseGraphParentMap;

bool get_tree(DiscretePairwiseGraph & g, DiscretePairwiseGraphColorMap color, DiscretePairwiseGraphParentMap parent)
{
	typedef DiscretePairwiseGraph Graph;
	
	typedef graph_traits <Graph>::vertex_descriptor VertexDescriptor;
	typedef property_traits<DiscretePairwiseGraphColorMap>::value_type ColorValue;
	typedef color_traits<ColorValue> Color;
	
	graph_traits <Graph>::vertex_iterator v, v_end;
	graph_traits <Graph>::adjacency_iterator u, u_end;
	
	priority_queue< GraphPartitionQueueOrder<Graph>::myTuple , vector < GraphPartitionQueueOrder<Graph>::myTuple >, GraphPartitionQueueOrder <Graph> > gray_vertices ( (GraphPartitionQueueOrder <Graph> (g)) );
	
	list <VertexDescriptor> backtrack_list_black;
	list <VertexDescriptor> backtrack_list_gray;
	
	vector <unsigned int> free_edges;

	bool trapped_vertex;
	unsigned int a;
	unsigned int b(0);
	
	VertexDescriptor vertex_with_lowest_degree = VertexDescriptor();

	//cout << "******************************************************" << endl;
	//cout << "We are starting the growing of a new tree." << endl;

	
	tie(v, v_end) = vertices(g);
	
	// If there are no more vertices, we are finished, return false
	
	if (v == v_end)
	{
		return false;
	}
	
	// Initialize all vertices to White, and they should be their own parents
	
	for (tie(v, v_end) = vertices(g); v != v_end; ++v)
	{
		put(color,*v, Color::white());
		put(parent,*v, *v);
	}
	
	//cout << "end init_tree()" << endl;
	
	if (prune_out_degree_0(g, color))
	{
		return true;
	}
	
	prune_out_degree_1(g, color, parent);
	
	prune_out_degree_2(g, color, parent);
	
	for (tie(v, v_end) = vertices(g); v != v_end; ++v)
	{
		if (get(color,*v) == Color::green())
		{
			continue;
		}
		
		vertex_with_lowest_degree = *v;
		b = out_degree(vertex_with_lowest_degree, g);
		break;
	}
	
	for (tie(v, v_end) = vertices(g); v != v_end; ++v)
	{
		free_edges.push_back(out_degree(*v, g));
		
		if (get(color,*v) == Color::green())
		{
			continue;
		}
		
		a = out_degree(*v, g);
		
		if (a < b)
		{
			b = a;
			vertex_with_lowest_degree = *v;
		}
	}
	
	
	vector<list <VertexDescriptor> > available_neighbors (num_vertices(g), list <VertexDescriptor> () );
	
	for (tie(v, v_end) = vertices(g); v != v_end; ++v)
	{
		for (tie (u, u_end) = adjacent_vertices(*v, g); u != u_end; ++u)
		{
			available_neighbors[*v].push_back(*u);
		}
	}
	
	// Obtain vertex with lowest degree
	
	gray_vertices.push( make_tuple (vertex_with_lowest_degree, compute_possible_choices(g, color, vertex_with_lowest_degree), out_degree(vertex_with_lowest_degree, g), 0, 0) );
	
	put(color, vertex_with_lowest_degree, Color::gray());
	
	unsigned step = 0;
	
	VertexDescriptor candidate;
	
	// Follow subroutine for a given subgraph
	
	while(!gray_vertices.empty())
	{
		
		// It is *only* a candidate because of the backtrack mechanism
		
		GraphPartitionQueueOrder<Graph>::myTuple top_element = gray_vertices.top();
		
		candidate = top_element.get<0>();
		
		//unsigned int old_choices = gray_vertices.top().get<1>();
		
		//unsigned int new_choices = compute_possible_choices(g, color, candidate);
		
		gray_vertices.pop();
		
		/*
		 if ( old_choices != new_choices)
		 {
			 cout << "Update: for " << g[candidate]->get_random_variable()->get_index() << " , we now have " << new_choices << " choices now." << endl;
			 
			 gray_vertices.push( make_tuple(candidate, new_choices, top_element.get<2>(), top_element.get<3>(), top_element.get<4>() ) );
			 
			 continue;
		 }
		 */
		//cout << "Staying at " << new_choices << " choices." << endl;
		
		
		if (get(color, candidate) != Color::gray())
		{
			//cerr << "Candidate " << g[candidate]->get_random_variable()->get_index() << " was refused " << endl;
			continue;
		}
		
		++step;
		trapped_vertex = false;
		
		// Loop on the neighbors exploration
		
		for (tie (u, u_end) = adjacent_vertices(candidate, g); u != u_end; ++u)
		{
			// Red means already chosen (in fact, the parent of the candidate)
			// We never do anything with the parent
			
			if ( get(color,*u) == Color::red() )
			{
				continue;
			}
			
			if (free_edges[*u] == 1)
			{
				trapped_vertex = true; 
			}
			
			// Black means already not possible, we still add it to the backtrack_list
			// This is because we need to decrease the free edges
			// Note that black and grey have the same effect
			
			if ( get(color,*u) == Color::black() )
			{
				backtrack_list_black.push_back(*u);
				continue;
			}
			
			if ( get(color,*u) == Color::gray())
			{
				backtrack_list_black.push_back(*u);
				continue;
			}
			
			// Here check if it has one neighbor that is black (or later)
			
			backtrack_list_gray.push_back(*u);
		}
		
		// If trapped_vertex is set to true, we have to backtrack. We undo everything and don't accept the candidate.
		
		if (trapped_vertex)
		{
			backtrack_list_black.clear();
			backtrack_list_gray.clear();
			//cout << ". Refused. "  << endl;
		}
		
		// Else candidate is accepted and we make all the modifications.
		
		else
		{
			//cout << ". Accepted. "  << endl;
			put(color, candidate, Color::red());
			
			for (list<VertexDescriptor>::iterator it = backtrack_list_black.begin() ; it != backtrack_list_black.end(); ++it)
			{
				put(color, *it, Color::black());
				free_edges [*it] = free_edges [*it] -1;
				available_neighbors[*it].remove(candidate);
				
				// If we reach only one free edge, we need to transmit (propagate) that !!
				
				if ( free_edges [*it] == 1)
				{
					//cout << g[*it]->get_random_variable()->get_index() << " has reached 1 and now we escape" << endl;
					
					transmit_escape(*it, color, free_edges, available_neighbors, g);
				}
				
			}
			
			for ( list<VertexDescriptor>::iterator it = backtrack_list_gray.begin() ; it != backtrack_list_gray.end(); ++it)
			{
				put(color, *it, Color::gray());
			}
			
			for ( list<VertexDescriptor>::iterator it = backtrack_list_gray.begin() ; it != backtrack_list_gray.end(); ++it)
			{
				// Following is old code not used anymore
				
				/*
				unsigned int quantity(0);
				
				if (check_black_neighbor(g, color, *it))
				{
					quantity = 10;
				}
					
					if (check_gray_neighbor(g, color, *it))
					{
						quantity = 5;
					}
				*/
				
				
				gray_vertices.push(make_tuple (*it, compute_possible_choices(g, color, *it), out_degree(*it, g), step, 0));
				available_neighbors[*it].remove(candidate);
				free_edges [*it] = free_edges [*it] -1;
			}
			
			backtrack_list_black.clear();
			backtrack_list_gray.clear();
		}
		
    }
	
	return true;
}


template <class Graph>
void partition_mrf_graph_into_chains (Graph & g)
{
	// Will work for graphs of size 2n (100*100...)
	
	const unsigned int top_left(0);
	const unsigned int bottom_left(1);
	const unsigned int bottom_right(2);
	const unsigned int top_right(3);
	
	// VERY UGLY
	
	unsigned int rows = mrf_rows;
	unsigned int columns = mrf_columns;
	
	// End hard coded
	
	typename graph_traits<Graph>::vertex_descriptor current_node = 0;
	
	unsigned int state = top_left;
	
	vector <bool> bool_vec(rows*columns, true);
	
	bool bool_cont = true;
	
	Graph & new_chain  = g.create_subgraph();
	
	add_vertex( current_node, new_chain);
	add_vertex( current_node+rows*columns, new_chain);
	bool_vec[current_node] = false;
	
	while (bool_cont)
	{
		
		switch (state)
		{
			case top_left:
			{
				
				++current_node;
				
				bool_cont = false;
				while (  (current_node % rows != 0) && bool_vec[current_node+1] )
				{
					bool_cont = true;
					add_vertex( current_node, new_chain);
					add_vertex( current_node+rows*columns, new_chain);
					bool_vec[current_node] = false;
					//cout << "Adding " << current_node << endl;
					++current_node;
				}
				
				--current_node;
				
				state = bottom_left;
			};
				
				break;
				
			case bottom_left:
			{
				current_node += rows;
				
				bool_cont = false;
				while (  current_node <  rows*(columns-1)  && bool_vec[current_node+rows] )
				{
					bool_cont = true;
					add_vertex( current_node, new_chain);
					add_vertex( current_node+rows*columns, new_chain);
					bool_vec[current_node] = false;
					//cout << "Adding " << current_node << endl;
					current_node += rows;
				}
				
				current_node -= rows;
				
				state = bottom_right;
			};
				
				break;
				
			case bottom_right:
			{
				--current_node;
				
				bool_cont = false;
				while ( (current_node %  rows) != 0 && bool_vec[current_node-1] )
				{
					bool_cont = true;
					add_vertex( current_node, new_chain);
					add_vertex( current_node+rows*columns, new_chain);
					bool_vec[current_node] = false;
					//cout << "Adding " << current_node << endl;
					--current_node;
				}
				
				++current_node;
				
				state = top_right;
			};
				break;
				
			case top_right:
			{
				current_node -= rows;
				
				bool_cont = false;
				while ( bool_vec[current_node-rows] )
				{
					bool_cont = true;
					add_vertex( current_node, new_chain);
					add_vertex( current_node+rows*columns, new_chain);
					bool_vec[current_node] = false;
					//cout << "Adding " << current_node << endl;
					current_node -= rows;
					//cout << "next 2 " << current_node-rows << ", " << bool_vec[current_node-rows] << endl;
				}
				
				current_node += rows;
				
				state = top_left;
			};
				break;
				
		}
	}
	
	//cout << "finished" << endl;
	
	Graph & second_chain = g.create_subgraph();
	
	for (unsigned int it = 0; it < bool_vec.size(); ++it)
	{
		if ( bool_vec[it])
		{
			add_vertex(it, second_chain);
			add_vertex(it+rows*columns, second_chain);
		}
	}	
}

template <class Graph>
void partition_simple_chain(Graph & g)
{
	typename graph_traits <Graph>::vertex_iterator u, u_end;
	
	Graph & chain = g.create_subgraph();
	
	for (tie (u, u_end) = vertices(g); u != u_end; ++u)
	{
		add_vertex(*u, chain);
	}
}

template <class Graph, class ColorMap, class ParentMap>
bool check_rv( Graph & g, typename graph_traits<Graph>::vertex_descriptor candidate, list < typename graph_traits<Graph>::vertex_descriptor> & backtrack_list_gray, ColorMap  & color, ParentMap & parent)
{
	typename graph_traits <Graph>::adjacency_iterator u, u_end;
	typedef typename property_traits<ColorMap>::value_type ColorValue;
	typedef color_traits<ColorValue> Color;
	
	for (tie (u, u_end) = adjacent_vertices(candidate, g); u != u_end; ++u)
	{
		if ( get(color,*u) == Color::red() )
		{
			continue;
		}
		
		if ( get(color, *u) == Color::black() )
		{
			//backtrack_list_black.push_back(*u);
			cout << "Something strange happened. It should never be black";
			continue;
		}
		
		if ( get(color, *u) == Color::gray())
		{
			//backtrack_list_black.push_back(*u);
			//This means we shouldn't choose this vertex at all
			return false;
		}
		
		//cout << "We grayed the Potential number " << g[*u]->index << endl; 
		
		backtrack_list_gray.push_back(*u);
	}
	
	return true;
}

template <class Graph, class ColorMap, class ParentMap>
bool	check_potential( Graph & g, typename  graph_traits<Graph>::vertex_descriptor candidate, list < typename graph_traits<Graph>::vertex_descriptor> & backtrack_list_gray, list < typename graph_traits<Graph>::vertex_descriptor> & backtrack_list_black, ColorMap  & color, ParentMap & parent, vector <unsigned int> & free_edges)
{
	typename graph_traits <Graph>::adjacency_iterator u, u_end;
	typedef typename property_traits<ColorMap>::value_type ColorValue;
	typedef color_traits<ColorValue> Color;
	
	bool non_trapped_vertex(true);
	
	for (tie (u, u_end) = adjacent_vertices(candidate, g); u != u_end; ++u)
	{
		if ( get(color,*u) == Color::red() )
		{
			continue;
		}
		
		if (free_edges[*u] == 1)
		{
			//non_trapped_vertex = false; 
		}
		
		if ( get(color,*u) == Color::black() )
		{
			//backtrack_list_black.push_back(*u);
			//cout << "Something strange happened (or not)" << endl;
			continue;
		}
		
		if ( get(color,*u) == Color::gray())
		{
			backtrack_list_black.push_back(*u);
			
			continue;
		}
		
		//cout << "We grayed the Variable number " << g[*u]->get_random_variable()->get_index() << endl; 
		
		backtrack_list_gray.push_back(*u);
	}
	
	return non_trapped_vertex;
}


typedef iterator_property_map < vector < default_color_type >::iterator, property_map < SubDiscreteFactorGraph, vertex_index_t>::type > DiscreteFactorGraphColorMap;
typedef iterator_property_map < vector < graph_traits <SubDiscreteFactorGraph>::vertex_descriptor >::iterator, property_map < SubDiscreteFactorGraph, vertex_index_t>::type > DiscreteFactorGraphParentMap;

bool get_phantom_potential(MessageNode < DiscreteRandomVariable, graph_traits <DiscreteFactorGraph>::vertex_descriptor> * uncasted )
{
	typedef graph_traits <DiscreteFactorGraph>::vertex_descriptor VertexDescriptor;
	
	PotentialMessageNode < DiscreteRandomVariable, DiscretePotential, VertexDescriptor > * ptr = dynamic_cast < PotentialMessageNode < DiscreteRandomVariable, DiscretePotential, VertexDescriptor > * > (uncasted);
	
	if (!ptr)
	{
		return false;
	}
	
	else
	{
		return ptr->get_phantom();
	}
}

//unsigned int teton = 0;

template <>
bool get_tree(DiscreteFactorGraph & g, DiscreteFactorGraphColorMap color, DiscreteFactorGraphParentMap parent)
{
	
	//cout << "We use the correct get_tree" << endl;
	
	typedef DiscreteFactorGraph Graph;
	
	typedef  graph_traits <Graph>::vertex_descriptor VertexDescriptor;
	typedef  property_traits<DiscreteFactorGraphColorMap>::value_type ColorValue;
	typedef color_traits<ColorValue> Color;
	
	graph_traits <Graph>::vertex_iterator v, v_end;
	
	priority_queue<GraphPartitionQueueOrder <Graph>::myTuple ,vector <GraphPartitionQueueOrder <Graph>::myTuple >, GraphPartitionQueueOrder <Graph> > gray_vertices ( (GraphPartitionQueueOrder <Graph> (g)) );
	
	graph_traits <Graph>::adjacency_iterator u, u_end;
	
	list <VertexDescriptor> backtrack_list_black;
	list <VertexDescriptor> backtrack_list_gray;
	
	vector <unsigned int> free_edges;
	
	unsigned int a(0);
	unsigned int b(0);
	
	VertexDescriptor vertex_with_lowest_degree = VertexDescriptor();
	
	// *********** CODE actually begins here ************ //
	
	//cout << "******************************************************" << endl;
	//cout << "We are starting the growing of a new tree." << endl;
	
	tie(v, v_end) = vertices(g);
	
	if (v == v_end)
	{
		return false;
	}
	
	//++teton;
	/*
	if (teton ==6)
	{
		exit (0);	
	}
	*/
	
	// Initialize ALL vertices to white, and they should be their OWN parents
	
	for (tie(v, v_end) = vertices(g); v != v_end; ++v)
	{
		put(color,*v, Color::white());
		put(parent,*v, *v);
	}
	
	/*
	 
	 I am not sure if we should do any sort of pruning for Factor Graphs
	 
	 prune_out_degree_1(g, color, parent);
	 
	 //cout << "end of pruning_1" << endl;
	 
	 prune_out_degree_2(g, color, parent);
	 
	 cout << "End of pruning_2" << endl;
	 
	 */
	
	for (tie(v, v_end) = vertices(g); v != v_end; ++v)
	{
		if (get(color,*v) == Color::green())
		{
			continue;
		}
		
		vertex_with_lowest_degree = *v;
		b = out_degree(vertex_with_lowest_degree, g);
		break;
	}
	
	for (tie(v, v_end) = vertices(g); v != v_end; ++v)
	{
		free_edges.push_back(out_degree(*v, g));
		
		if (get(color,*v) == Color::green())
		{
			continue;
		}
		
		a = out_degree(*v, g);
		
		if ( (a < b) && (!get_phantom_potential(g[*v])) )
		{
			b = a;
			vertex_with_lowest_degree = *v;
		}
		
	}
	
	// Obtain vertex with lowest degree
	
	gray_vertices.push(make_tuple (vertex_with_lowest_degree, compute_possible_choices(g, color, vertex_with_lowest_degree), out_degree(vertex_with_lowest_degree, g), 0, 0));
	
	put(color, vertex_with_lowest_degree, Color::gray());
	
	unsigned step = 0;
	
	VertexDescriptor candidate;
	
	// Follow subroutine for a given subgraph
	
	//cout << "Entering main loop" << endl;
	
	while(!gray_vertices.empty())
	{
		++step;
		
		bool non_trapped_vertex(false);
		
		// It is *only* a candidate because of the backtrack mechanism
		
		candidate = gray_vertices.top().get<0>();
		
		gray_vertices.pop();
		
		// Here we test if it is still a candidate, e.g. if it hasn't been pruned out already
		
		if (get(color, candidate) != Color::gray())
		{
			//cerr << "Candidate " << candidate+1 << " was refused " << endl;
			
			continue;
		}
		
		if (get_phantom_potential(g[candidate]))
		{
			//cerr << "Candidate was refused as a phantom potential." << endl;
			//exit(0);
			
			continue;
		}
		
		VariableMessageNode < DiscreteRandomVariable, DiscretePotential, VertexDescriptor> * ptr = dynamic_cast < VariableMessageNode < DiscreteRandomVariable, DiscretePotential, VertexDescriptor> * > (g [candidate]);
		
		if (ptr)
		{
			//cout << "We explore variable " << ptr->get_random_variable()->get_index() << endl;
			non_trapped_vertex = check_rv(g, candidate, backtrack_list_gray, color, parent);
		}
		
		else
		{
			//PotentialMessageNode < DiscreteRandomVariable, DiscretePotential, VertexDescriptor> * ptr = dynamic_cast < PotentialMessageNode < DiscreteRandomVariable, DiscretePotential, VertexDescriptor> * > (g [candidate]);
			
			//cout << "We explore potential " << g [candidate]->index << endl;
			non_trapped_vertex = check_potential (g, candidate, backtrack_list_gray, backtrack_list_black, color, parent, free_edges);
		}
		
		
		if (!non_trapped_vertex) // nothing happens... we don't choose this vertex
		{
			//cout << "Backtracking." << endl;
			backtrack_list_black.clear();
			backtrack_list_gray.clear();
			//put(color, candidate, Color::black());
		}
		
		else // incorporate backtrack stuff into main	
		{
			if (ptr)
			{
				//cout << "We added vertex " << ptr->get_random_variable()->get_index() << endl;
			}
			
			else
				
			{
				//cout << "We added potential " << g [candidate]->index << endl;
			}
			
			put(color, candidate, Color::red());
			
			for ( list<VertexDescriptor>::iterator it = backtrack_list_black.begin() ; it != backtrack_list_black.end(); ++it)
			{
				put(color, *it, Color::black());
				free_edges [*it] = free_edges [*it] -1;
			}
			
			for ( list<VertexDescriptor>::iterator it = backtrack_list_gray.begin() ; it != backtrack_list_gray.end(); ++it)
			{
				put(color, *it, Color::gray());
				
				gray_vertices.push(make_tuple (*it,  compute_possible_choices(g,  color, *it), out_degree(*it, g), step, 0));
				
				free_edges [*it] = free_edges [*it] -1;
			}
			
			backtrack_list_black.clear();
			backtrack_list_gray.clear();
		}
		
    }
	
	return true;
}

template <>
inline unsigned int check_phantom_potential_status(boost::graph_traits<SubDiscretePairwiseGraph>::vertex_descriptor , DiscretePairwiseGraphColorMap &, SubDiscretePairwiseGraph &)
{
	return 0;
}

template <>
unsigned int check_phantom_potential_status(boost::graph_traits<SubDiscreteFactorGraph>::vertex_descriptor u, DiscreteFactorGraphColorMap & color, SubDiscreteFactorGraph & g)
{
	typedef graph_traits <DiscreteFactorGraph>::vertex_descriptor VertexDescriptor;

	PotentialMessageNode < DiscreteRandomVariable, DiscretePotential, VertexDescriptor> * ptr = dynamic_cast < PotentialMessageNode < DiscreteRandomVariable, DiscretePotential, VertexDescriptor> * > (g [u]);
	
	if (!ptr)
	{
		return 0;
	}
	
	graph_traits <DiscreteFactorGraph>::adjacency_iterator v, v_end;
	typedef property_traits<DiscreteFactorGraphColorMap>::value_type ColorValue;
	typedef color_traits<ColorValue> Color;
	
	if ( (get(color,u) != Color::red() ) &&  (!ptr->phantom_potential ))
	{
		return 0;
	} 
	
	bool initial_status = ptr->phantom_potential;
	
	ptr->phantom_potential = false;
	
	unsigned int black_neighbors (0);
	
	//cout << "Examining potential " << ptr->index << " for phantom. ";
	
	for (tie (v, v_end) = adjacent_vertices(u, g); v != v_end; ++v)
	{
		//cout << "Neighbour " << g[*v]->get_random_variable()->get_index();
		
		if ( get(color,*v) == Color::red() )
		{
			//cout << ". Color is red." << endl;
			continue;
		} 
		

		if ( get(color,*v) == Color::black() )
		{
			//cout << ". Color is black." << endl;
		} 
		
		if ( get(color,*v) == Color::gray() )
		{
			//cout << ". Color is gray." << endl;
		} 
		
		if ( get(color,*v) == Color::white() )
		{
			//cout << ". Color is white." << endl;
		} 
		
		//cout << "Error: we encountered a strange color." << endl;
		
		++black_neighbors;
	}
	
	if ( black_neighbors > 1)
	{
		ptr->phantom_potential = true;
	}
	
	if (ptr->phantom_potential )
	{
		//cout << "Phantom." << endl;		
	}
	
	else
	{
		//cout << "Refused." << endl;	
	}
	
	if ((!ptr->phantom_potential) && (initial_status))
	{
		return 1;
	}
	
	if ((!ptr->phantom_potential) && (!initial_status))
	{
		cout << "Strange " << endl;
		return 0;
	}	
	
	if ((ptr->phantom_potential) && (!initial_status))
	{
		assert (get(color,u) == Color::red());
		return 2;
	}
		
	if ((ptr->phantom_potential) && (initial_status))
	{
		return 3;
	}
	
	cout << "FATAL ERROR" << endl;
		
	return 4;
}

template <class SubGraph, class ColorMap, class ParentMap>
void partition_graph_into_trees_implementation (SubGraph & g, ColorMap color, ParentMap parent)
{	
	typedef typename SubGraph::graph_type Graph;
	
	typedef typename property_traits<ColorMap>::value_type ColorValue;
	typedef color_traits<ColorValue> Color;
	
	typename graph_traits <Graph>::vertex_iterator u, u_end;
	typename graph_traits <Graph>::adjacency_iterator v, v_end;
	
	
	cout << "Starting the partition... " << flush;
	
	SubGraph * root_subgraph = new SubGraph();
	copy_graph(g,*root_subgraph);
	
	Graph * working_graph = new Graph();
	copy_graph(*root_subgraph,*working_graph);
	
	SubGraph & working_subgraph = root_subgraph->create_subgraph();
	
	working_subgraph = *root_subgraph; // assignment
	
	// The working graph is a GRAPH, not a subgraph. We make all the color operations on this graph.
	// We create it at each iteration.
	
	while ( get_tree(*working_graph, color, parent) )
	{
		
		// New tree will only be used to contain the correct stuff. It is directly linked to the parent graph as one of the subgraph.
		// We need not to bother with it.
		
		SubGraph & new_tree = g.create_subgraph();
		
		// The root subgraph will contain 
		
		SubGraph & new_working_subgraph = root_subgraph->create_subgraph();
		
		for (tie(u, u_end) = vertices(working_subgraph); u != u_end; ++u)
		{
			//cout << "nago" << endl;
			
			switch (check_phantom_potential_status(*u, color, working_subgraph))
			{
				case 0: break;
					
				case 1: // the potential was a phantom potential, but has just become 'normal'
				{
					//cout << "ARA" << endl;
					continue;
				}
					break;
					
				case 2: // the potential has just become a phantom potential
				{
					//cout << "ARA" << endl;
					
					add_vertex( working_subgraph.local_to_global(*u), new_tree);
					add_vertex( working_subgraph.local_to_global(*u), new_working_subgraph);
					continue;
				}
					break;
					
				case 3: // the potential was a phantom potential and continues to be a phantom potential
				{
					add_vertex( working_subgraph.local_to_global(*u), new_working_subgraph);
					continue;
				}	
					break;
					
				default: break;
			}
					
			//cout << "bobo" << endl;
			
			if (get (color,*u) == Color::red() )
			{
				//cout << "We encountered red vertex " << *u+1 << " ( " << working_subgraph.local_to_global(*u) + 1<< " )" << endl;
				add_vertex( working_subgraph.local_to_global(*u), new_tree);
				continue;
			}
			
			if (get (color,*u) == Color::green() )
			{
				//cout << "We encoutered green vertex " << *u+1 << " ( " << working_subgraph.local_to_global(*u) + 1<< " )" << endl;
				
				// Here we should test if the parent is Red Or Not
				
				if (get( color, get(parent, *u)) == Color::red() )
				{
					//cout << "Adding it because parent is Red" << endl;
					add_vertex( working_subgraph.local_to_global(*u), new_tree);
					continue;
				}
				
			}
			
			// Else...
			
			add_vertex( working_subgraph.local_to_global(*u), new_working_subgraph);
		}
		
		// Clearing the Graph
		
		delete working_graph;
		working_graph = new Graph();
		
		// Here the operations are very important, we create a new working subgraph, and make it identical to 
		
		copy_graph(new_working_subgraph, *working_graph);
		
		working_subgraph = new_working_subgraph; // trying assignment operator
	}
	
	delete working_graph;
	delete root_subgraph;
	
	//display_subgraph(g);
	
	typename subgraph<Graph>::children_iterator s, s_end;	
	
	unsigned int total_vertices = 0;
	
	for ( tie(s, s_end) = g.children(); s != s_end; ++s)
	{
		total_vertices += num_vertices(*s);
	}
	
	assert (total_vertices == num_vertices(g));
	
	cout << "We are finished with the partition." << endl;
}

template <class Graph>
void partition_graph_into_trees (Graph & g)
{
	vector <typename graph_traits <Graph>::vertex_descriptor> v_1 (num_vertices(g));
	vector <default_color_type> v_2 (num_vertices(g));
	
	partition_graph_into_trees_implementation(g, make_iterator_property_map(v_2.begin(), get(vertex_index,g)), make_iterator_property_map( v_1.begin(), get(vertex_index,g)));
}



